﻿#include "AuthHeaderParser.h"
#include "StrUtils.h"

CAuthHeaderParser::CAuthHeaderParser(const char* rawHeader) {
  parseRawHeader(rawHeader);
}

void CAuthHeaderParser::parseRawHeader(const std::string& rawHeader) {
  auto methodAndParams = CStrUtils::split(rawHeader, " ", 1);
  _method = methodAndParams[0];
  if (methodAndParams.size() > 1) {
    parseParams(methodAndParams[1]);
  }
}

void CAuthHeaderParser::parseParams(const std::string& paramStr) {
  auto params = CStrUtils::splitWithLiteral(paramStr, ",");
  for (auto param : params) {
    auto nameAndValue = CStrUtils::split(param, "=", 1);
    if (nameAndValue.size() > 1) {
      addParam(CStrUtils::trim(nameAndValue[0]), CStrUtils::trim(nameAndValue[1]));
    }
  }
}

void CAuthHeaderParser::addParam(const std::string& name, const std::string& value) {
  _params.push_back(struct ParamInfo { name, unquote(value) });
}

bool CAuthHeaderParser::hasMethod(const char* method) const {
  return method != nullptr &&
    CStrUtils::isEqualNoCase(_method, method);
}

bool CAuthHeaderParser::hasParam(const char* name, const char* value) const {
  auto info = getParamInfo(name);
  bool result = info != nullptr;
  if (result && value != nullptr) {
    result = CStrUtils::isEqualNoCase(info->value, value);
  }
  return result;
}

const char* CAuthHeaderParser::getParam(const char* name) const {
  auto info = getParamInfo(name);
  return (info != nullptr) ? info->value.c_str() : nullptr;
}

const struct CAuthHeaderParser::ParamInfo* CAuthHeaderParser::getParamInfo(const std::string& name) const {
  const struct ParamInfo* result = nullptr;
  for (auto i = _params.begin(); i != _params.end() && result == nullptr; i++) {
    if (CStrUtils::isEqualNoCase(i->name, name)) {
      result = &*i;
    }
  }
  return result;
}

std::string CAuthHeaderParser::unquote(const std::string& str) {
  return CStrUtils::trim(str, "\"");
}
