﻿#include <string.h>
#include "StrUtils.h"

#ifndef _strncasecmp
#define _strncasecmp _strnicmp
#endif

CStrUtils::CCharIterator::CCharIterator(const std::string& str, unsigned int tailLength) {
    _str = str;
    _tailLength = tailLength;
    moveFirst();
}

bool CStrUtils::CCharIterator::hasNext() const {
    return _str.size() > _tailLength + 1 && _current < _str.size() - 1 - _tailLength;
}

unsigned int CStrUtils::CCharIterator::moveFirst() {
    _current = 0;
    return moveNextReadable();
}

unsigned int CStrUtils::CCharIterator::moveNext() {
    if (hasNext()) {
        _current++;
    }
    return moveNextReadable();
}

unsigned int CStrUtils::CCharIterator::moveNextReadable() {
    return _current;
}

CStrUtils::CLiteralIterator::CLiteralIterator(const std::string& str, unsigned int tailLength): CCharIterator(str, tailLength) {
}

unsigned int CStrUtils::CLiteralIterator::moveFirst() {
    _escapeChar = '\\';
    _literalBoundChar = '\"';
    _escaped = false;
    _inLiteral = false;
    return CCharIterator::moveFirst();
}

unsigned int CStrUtils::CLiteralIterator::moveNextReadable() {
    for (; hasNext(); _current++) {
        if (_escaped) {
            _escaped = false;
        }
        else if (_str[_current] == _escapeChar) {
            _escaped = true;
        }
        else {
            if (_str[_current] == _literalBoundChar) {
                _inLiteral = !_inLiteral;
            }
            else if (!_inLiteral) {
                break;
            }
        }
    }
    return _current;
}

std::vector<std::string> CStrUtils::split(const std::string& str, const std::string& separator, int maxParts) {
    CCharIterator iterator(str, separator.size());
    return splitOnIterator(str, separator, maxParts, &iterator);
}

std::vector<std::string> CStrUtils::splitWithLiteral(const std::string& str, const std::string& separator, int maxParts) {
    CLiteralIterator iterator(str, separator.size());
    return splitOnIterator(str, separator, maxParts, &iterator);
}

std::vector<std::string> CStrUtils::splitOnIterator(const std::string& str, const std::string& separator, int maxParts, CCharIterator* iterator) {
    std::vector<std::string> result;
    unsigned int startPos = iterator->moveFirst();
    for (unsigned int i = startPos; iterator->hasNext() && maxParts != 0; i = iterator->moveNext()) {
        if (str.substr(i, separator.size()) == separator) {
            result.push_back(str.substr(startPos, i - startPos));
            startPos = i += separator.size();
            maxParts--;
        }
    }
    result.push_back(str.substr(startPos,  str.size() - startPos));
    return result;
}

bool CStrUtils::isEqualNoCase(const std::string& str1, const std::string& str2) {
    return str1.size() == str2.size() &&
        _strncasecmp(str1.c_str(), str2.c_str(), str2.size()) == 0;
}

std::string CStrUtils::trim(const std::string& str, const std::string& trimSymbols) {
    return trimUni(trimUni(str, trimSymbols, str.size() - 1, -1), trimSymbols, 0, 1);
}

std::string CStrUtils::trimUni(const std::string& str, const std::string& trimSymbols, unsigned int initPos, int increment) {
    std::string result;
    unsigned int noTrimSymbol = initPos;
    for (; noTrimSymbol >= 0 && noTrimSymbol < str.size(); noTrimSymbol+=increment) {
        if (trimSymbols.find(str[noTrimSymbol]) == std::string::npos) {
            break;
        }
    }
    if (noTrimSymbol >= 0 && noTrimSymbol < str.size()) {
        if (increment > 0) {
            result = str.substr(noTrimSymbol);
        }
        else {
            result = str.substr(0, noTrimSymbol + 1);
        }
    }
    return result;
}


