#pragma once

#include <string>
#include <vector>

class CStrUtils {
private:
    class CCharIterator {
    public:
        CCharIterator(const std::string& str, unsigned int tailLength = 0);

        bool hasNext() const;
        virtual unsigned int moveFirst();
        unsigned int moveNext();
    protected:
        unsigned int _current;
        std::string _str;
        unsigned int _tailLength;

        virtual unsigned int moveNextReadable();
    };

    class CLiteralIterator: public CCharIterator {
    public:
        CLiteralIterator(const std::string& str, unsigned int tailLength = 0);

        unsigned int moveFirst();
    protected:
        bool _escaped;
        bool _inLiteral;
        char _escapeChar;
        char _literalBoundChar;

        unsigned int moveNextReadable();
    };

    static std::string trimUni(const std::string& str, const std::string& trimSymbols, unsigned int initPos, int increment);
    static std::vector<std::string> splitOnIterator(const std::string& str, const std::string& separator, int maxParts, CCharIterator* iterator);
public:
    static std::vector<std::string> split(const std::string& str, const std::string& separator, int maxParts = -1);
    static std::vector<std::string> splitWithLiteral(const std::string& str, const std::string& separator, int maxParts = -1);
    static bool isEqualNoCase(const std::string& str1, const std::string& str2);
    static std::string trim(const std::string& str, const std::string& trimSymbols = std::string(" "));
};