﻿#include "Strings.h"
#include <algorithm>
#include <regex>

std::vector<std::string> CStrings::split(const std::string& str, const std::string& separator)
{
  std::vector<std::string> separators;
  separators.push_back(separator);
  return split(str, separators);
}

bool isInString(const std::string& str, std::string::size_type point, const std::string& separator)
{
  bool inString = false;
  std::string::size_type remainStrLen = str.size() - point;
  if (remainStrLen >= separator.size()) {
    inString = true;
    for (std::string::size_type i = 0; i < separator.size() && inString; i++) {
      inString = (str[point + i] == separator[i]);
    }
  }
  return inString;
}

std::vector<std::string>::const_iterator startWithAny(const std::string& str, std::string::size_type point, const std::vector<std::string>& separators)
{
  auto separator = separators.begin();
  do {
    if (isInString(str, point, *separator)) {
      break;
    }
    separator++;
  } while (separator != separators.end());
  return separator;
}

std::vector<std::string> CStrings::split(const std::string& str, const std::vector<std::string>& separators)
{
  std::string::size_type previousPoint = 0;
  std::vector<std::string> result;
  for (auto point = previousPoint; point < str.size(); point++) {
    bool isTheLastSymbol = (point + 1) >= str.size();
    auto foundSeparator = startWithAny(str, point, separators);
    if (foundSeparator != separators.end() || isTheLastSymbol) {
      result.push_back(str.substr(previousPoint, point - previousPoint + (isTheLastSymbol ? 1 : 0)));
    }
    if (foundSeparator != separators.end()) {
      point += (*foundSeparator).size() - 1;
      previousPoint = point + 1;
    }
  }
  return result;
}

std::vector<std::string> CStrings::lookForPropertyValues(std::vector<std::string> strings, const std::string& startSequence, int maxOccurens)
{
  auto found = findLinesStartsWith(strings, startSequence, maxOccurens);
  std::transform(found.begin(), found.end(), found.begin(), [startSequence](const std::string& elem) -> std::string {
    std::string copy(elem);
    copy.erase(0, startSequence.size());
    return copy;
  });
  return found;
}

std::vector<std::string> CStrings::findLinesStartsWith(std::vector<std::string> strings, const std::string& startSequence, int maxOccurens)
{
  std::vector<std::string> result;
  auto current = strings.begin();
  while (current != strings.end() && maxOccurens != result.size()) {
    current = std::find_if(current, strings.end(), [startSequence](const std::string& elem) -> bool {
      return elem.find(startSequence) == 0;
    });
    if (current != strings.end()) {
      result.push_back(*current);
      current++;
    }
  }
  return result;
}

std::string CStrings::trim(const std::string& str)
{
  return trimRight(trimLeft(str));
}

std::string clearWithRegex(const std::string& str, const std::string& regExPattern)
{
  std::regex pattern(regExPattern);
  return std::regex_replace(str, pattern, "");
}

std::string CStrings::trimLeft(const std::string& str)
{
  return clearWithRegex(str, "^(\\s|\\n|\\r)*");
}

std::string CStrings::trimRight(const std::string& str)
{
  return clearWithRegex(str, "(\\s|\\n|\\r)*$");
}

std::string stringToLower(const std::string& str)
{
  std::string result(str);
  for (std::string::size_type i = 0; i < result.size(); i++) {
    result[i] = tolower(result[i]);
  }
  return result;
}

bool CStrings::includesNoCase(const std::string& str, const std::string& part)
{
  return findNoCase(str, part) != std::string::npos;
}

bool CStrings::includes(const std::string & str, const std::string & part)
{
	return str.find(part) != std::string::npos;
}

bool CStrings::equalsNoCase(const std::string& str1, const std::string& str2)
{
  return str1.size() == str2.size() &&
    startsWithNoCase(str1, str2);
}

std::string::size_type CStrings::findNoCase(const std::string& str, const std::string& substring)
{
  return stringToLower(str).find(stringToLower(substring));
}

bool CStrings::startsWithNoCase(const std::string& str, const std::string& substring)
{
  return str.size() >= substring.size() &&
    0 == _strnicmp(str.c_str(), substring.c_str(), substring.size());
}

bool CStrings::endsWith(const std::string & str, const std::string & substring)
{
	return (str.size() > substring.size()) ? isInString(str, str.size() - substring.size(), substring) : false;
}

std::string CStrings::ensureEndsWith(const std::string & str, const std::string & end)
{
  return endsWith(str, end) ? str : (str + end);
}

bool CStrings::NoCaseComparer::operator() (const std::string& str1, const std::string& str2) const
{
  return stringToLower(str1) < stringToLower(str2);
}
